/* countdown - count down a specified time interval
**
** Copyright  2012,2014,2015,2022 by Jef Poskanzer <jef@mail.acme.com>.
** All rights reserved.
**
** You may use this software however you like as long as you keep my
** name on it and don't sue me.
**
** For commentary on this license please see http://acme.com/license.html
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <ctype.h>

#define MINUTE 60
#define HOUR (60*MINUTE)
#define DAY (24*HOUR)
#define WEEK (7*DAY)


static char* argv0;


static void usage( void )  __attribute__ ((noreturn));
static double time_f( void );


int
main( int argc, char** argv )
    {
    char* interval;
    size_t interval_len;
    int scan_len;
    double now, end;
    time_t t;
    struct tm tm;
    char ap, u1, u2;
    double d1, d2;
    int i1, i2, i3;
    int mul1, mul2;
    long seconds, prev_seconds;
    long d, h, m, s;
    int i;
    char* pad = "          ";

    argv0 = argv[0];
    if ( argc != 2 )
	usage();
    interval = argv[1];
    interval_len = strlen( argv[1] );

    /* Parse the interval */
    now = time_f();
    t = time( (time_t*) 0 );
    (void) localtime_r( &t, &tm );
    if ( sscanf( interval, "%d:%d:%d%cm%n", &i1, &i2, &i3, &ap, &scan_len ) == 4 && scan_len == interval_len )
	{
	ap = tolower( ap );
	if ( ap != 'a' && ap != 'p' )
	    usage();
	tm.tm_hour = i1;
	if ( ap == 'p' && tm.tm_hour < 12 )
	    tm.tm_hour += 12;
	tm.tm_min = i2;
	tm.tm_sec = i3;
	end = mktime( &tm );
	while ( end < now )
	    end += DAY;
	}
    else if ( sscanf( interval, "%d:%d:%d%n", &i1, &i2, &i3, &scan_len ) == 3 && scan_len == interval_len )
	{
	tm.tm_hour = i1;
	tm.tm_min = i2;
	tm.tm_sec = i3;
	end = mktime( &tm );
	while ( end < now )
	    {
	    if ( i1 > 12 )
		end += DAY;
	    else
		end += 12 * HOUR;
	    }
	}
    else if ( sscanf( interval, "%d:%d%cm%n", &i1, &i2, &ap, &scan_len ) == 3 && scan_len == interval_len )
	{
	ap = tolower( ap );
	if ( ap != 'a' && ap != 'p' )
	    usage();
	tm.tm_hour = i1;
	if ( ap == 'p' && tm.tm_hour < 12 )
	    tm.tm_hour += 12;
	tm.tm_min = i2;
	tm.tm_sec = 0;
	end = mktime( &tm );
	while ( end < now )
	    end += DAY;
	}
    else if ( sscanf( interval, "%d:%d%n", &i1, &i2, &scan_len ) == 2 && scan_len == interval_len )
	{
	tm.tm_hour = i1;
	tm.tm_min = i2;
	tm.tm_sec = 0;
	end = mktime( &tm );
	while ( end < now )
	    {
	    if ( i1 > 12 )
		end += DAY;
	    else
		end += 12 * HOUR;
	    }
	}
    else if ( sscanf( interval, "%lf%c%lf%c%n", &d1, &u1, &d2, &u2, &scan_len ) == 4 && scan_len == interval_len )
	{
	u1 = tolower( u1 );
	u2 = tolower( u2 );
	switch ( u1 )
	    {
	    case 's': mul1 = 1; break;
	    case 'm': mul1 = MINUTE; break;
	    case 'h': mul1 = HOUR; break;
	    case 'd': mul1 = DAY; break;
	    case 'w': mul1 = WEEK; break;
	    default: usage();
	    }
	switch ( u2 )
	    {
	    case 's': mul2 = 1; break;
	    case 'm': mul2 = MINUTE; break;
	    case 'h': mul2 = HOUR; break;
	    case 'd': mul2 = DAY; break;
	    case 'w': mul2 = WEEK; break;
	    default: usage();
	    }
	end = now + d1 * mul1 + d2 * mul2;
	}
    else if ( sscanf( interval, "%lf%c%n", &d1, &u1, &scan_len ) == 2 && scan_len == interval_len )
	{
	u1 = tolower( u1 );
	switch ( u1 )
	    {
	    case 's': mul1 = 1; break;
	    case 'm': mul1 = MINUTE; break;
	    case 'h': mul1 = HOUR; break;
	    case 'd': mul1 = DAY; break;
	    case 'w': mul1 = WEEK; break;
	    default: usage();
	    }
	end = now + d1 * mul1;
	}
    else if ( sscanf( interval, "%lf%n", &d1, &scan_len ) == 1 && scan_len == interval_len )
	end = now + d1;
    else
	usage();

    /* Do the countdown. */
    prev_seconds = -1;
    for (;;)
	{
	now = time_f();
	seconds = ( end - now ) + 0.5;
	d = seconds / DAY;
	h = ( seconds % DAY ) / HOUR;
	m = ( seconds % HOUR ) / MINUTE;
	s = seconds % MINUTE;
	if ( seconds != prev_seconds )
	    {
	    if ( d > 0 )
		{
		if ( s > 0 )
		    (void) printf( "\r%ldd %ldh %ldm %lds%s", d, h, m, s, pad );
		else
		    {
		    if ( m > 0 )
			(void) printf( "\r%ldd %ldh %ldm%s", d, h, m, pad );
		    else
			{
			if ( h > 0 )
			    (void) printf( "\r%ldd %ldh%s", d, h, pad );
			else
			    (void) printf( "\r%ldd%s", d, pad );
			}
		    }
		}
	    else if ( h > 0 )
		{
		if ( s > 0 )
		    (void) printf( "\r%ldh %ldm %lds%s", h, m, s, pad );
		else
		    {
		    if ( m > 0 )
			(void) printf( "\r%ldh %ldm%s", h, m, pad );
		    else
			(void) printf( "\r%ldh%s", h, pad );
		    }
		}
	    else if ( m > 0 )
		{
		if ( s > 0 )
		    (void) printf( "\r%ldm %lds%s", m, s, pad );
		else
		    (void) printf( "\r%ldm%s", m, pad );
		}
	    else
		(void) printf( "\r%lds%s", s, pad );
	    (void) fflush( stdout );
	    if ( now >= end || seconds <= 0 )
		break;
	    prev_seconds = seconds;
	    }
	usleep( 200000 );
	}

    /* Alarm. */
    (void) printf( "\n" );
    for ( i = 0; i < 5; ++i )
	{
	(void) putchar( '' );
	(void) fflush( stdout );
	sleep( 1 );
	}

    exit( EXIT_SUCCESS );
    }


static void
usage( void )
    {
    (void) fprintf( stderr, "usage:  %s interval\n", argv0 );
    exit( EXIT_FAILURE );
    }


static double
time_f( void )
    {
    struct timeval tv;

    (void) gettimeofday( &tv, NULL );
    return (double) tv.tv_sec + (double) tv.tv_usec / 1000000.0;
    }
